// -----------------------------------------------------------
// ClientSocket.java
// -----------------------------------------------------------


package cz.vutbr.wsagent.client;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;

/**
 * Use for creating new Socket to connect to the BaseStation software.
 * After successful connection is able to send message.
 *
 * @author Martin Gábor.
 */
public class ClientSocket {

    // socket
    private Socket socket = null;


    /**
     * Constructor for initializing new socket. Localhost is a default address
     * and it can not be changed outside this class.
     *
     * @param port Socket port of socket.
     * @exception IOException Signals that an I/O exception while getting local host
     *            has occurred.
     * @exception UnknownHostException Thrown if host is inaccessible.
     */
    public ClientSocket(int port) throws UnknownHostException, IOException {

        // LocalHost is used as an address of receiver
        InetAddress host = InetAddress.getLocalHost();

        // init socket
        socket = new Socket(host.getHostName(), port);
    }


    /**
     * This method is used to send messages using socket. Socket itself
     * describes the communication properties. Communication is provided by
     * <code>ObjectOutputStream</code> and <code>ObjectInpuputStream</code>. The
     * second one is used for receiving result from host. If the result is not SENDING_SUCCESS
     * new CommunicationErrorException is thrown.
     *
     * @param inputMessage Message to send.
     * @exception ClassNotFoundException Thrown when an application tries to load in a
     *            class through its string name.
     * @exception IOException Signals that an I/O exception while working with socket has occurred.
     * @exception CommunicationErrorException Thrown when incoming result describes
     *            unsuccessful operation.
     * @exception TimeExpiredException Thrown when a communicating time expires.
     * @exception DeviceBusyException Thrown when the device is busy.
     * @see ClientSocket
     *
     */
    public void send(String inputMessage) throws IOException, ClassNotFoundException, 
            CommunicationErrorException, TimeExpiredException, DeviceBusyException {

        // sending
        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket.getOutputStream());
        objectOutputStream.writeObject(inputMessage);

        // reading
        ObjectInputStream objectInputStream = new ObjectInputStream(socket.getInputStream());
        String message = null;

        message = (String) objectInputStream.readObject();
            

        // closing
        objectOutputStream.close();
        objectInputStream.close();


        // handle result of process
        if(message.compareTo(Integer.toString(InternalCommProtocol.SENDING_ERROR)) == 0)
            throw new CommunicationErrorException();
        if(message.compareTo(Integer.toString(InternalCommProtocol.SENDING_EXPIRED)) == 0)
            throw new TimeExpiredException();
        if(message.compareTo(Integer.toString(InternalCommProtocol.SENDING_DEV_BUSY)) == 0)
            throw new DeviceBusyException();
    }


    /**
     * This method closes the socket.
     *
     * @exception IOException Signals that an I/O exception while working with socket
     *            has occurred.
     */
    public void close() throws IOException{
        socket.close();
    }

}
