// -----------------------------------------------------------
// WSAgentSend.java
// -----------------------------------------------------------
package cz.vutbr.wsagent.controlpanel;

import cz.vutbr.wsagent.client.DeviceBusyException;
import cz.vutbr.wsagent.client.TimeExpiredException;
import cz.vutbr.wsagent.controlpanel.exception.SendingMessageException;
import cz.vutbr.wsagent.controlpanel.exception.InvalidParametersException;
import cz.vutbr.wsagent.controlpanel.exception.DatabaseHandlingException;
import cz.vutbr.wsagent.db.hibernate.DatabaseErrorException;
import cz.vutbr.wsagent.client.ClientSocket;
import cz.vutbr.wsagent.client.CommunicationErrorException;
import cz.vutbr.wsagent.client.MessageConstructor;
import cz.vutbr.wsagent.db.hibernate.DAOFactory;
import cz.vutbr.wsagent.entity.Agent;
import java.io.IOException;
import java.net.ConnectException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.Date;

/**
 * This action class is used for handling agent's data. It cand send it to BSComm or save it
 * to the database.
 *
 * @author Martin Gábor
 */
public class WSAgentSend extends ControlPanelAction {

    private List<Agent> agents;
    private Agent agentBean;
    private String address;
    private String param;
    private long aId;

    /**
     * Execute Method.
     */
    @Override
    public String execute() {

        initAgentList();

        // get all saved agents
        try {
            agents = DAOFactory.getDAOFactory().getAgentDAO().readAll();
        } catch (DatabaseErrorException ex) {
        }
        setAgents(agents);

        // if something has been submitted
        if (param != null) {
            
            // if it is Send
            if (param.compareTo("Send") == 0) {
                setResponseMessage(getText("Comm.Success.agentsend"));
                try {
                    // send.
                    send();
                } catch (InvalidParametersException ex) {
                    setResponseMessage(getText("InvalidParameters.message"));
                    return INPUT;
                } catch (ConnectException ex) {
                    setResponseMessage(getText("Comm.ConnectError.message"));
                } catch (SendingMessageException ex) {
                    setResponseMessage(getText("Comm.SendingError.message"));
                } catch (TimeExpiredException ex) {
                    setResponseMessage(getText("Comm.TimeExpired.message"));
                } catch (DeviceBusyException ex) {
                    setResponseMessage(getText("Comm.DeviceBusy.message"));
                }
                try {
                    // now if is agent sended - change because tracing agent
                    // agent is inserted in database
                    savesend();
                } catch (DatabaseHandlingException ex) {
                    setResponseMessage(getText("Db.Error.message"));
                }
                
                // if it is Save
            } else {
                setResponseMessage(getText("Db.Success.message"));
                try {
                    // save
                    save();
                } catch (DatabaseHandlingException ex) {
                    setResponseMessage(getText("Db.Error.message"));
                }
            }
        }

        // operation Success
        return SUCCESS;
    }

    /**
     * Checks if given value is invalid.
     * @param value Value to be validated.
     * @return Returns true if the value is invalid
     */
    private boolean isInvalid(String value) {
        return (value == null || value.length() == 0);
    }

    /**
     * Inits the arrayList of agents.
     */
    private void initAgentList() {
        if (agents == null) {
            agents = new ArrayList();
        }
    }

    /**
     * This method handles submited data to be send and creates socket for sending data to
     * BSComm application.
     *
     * @throws ConnectException Thrown if can not connect to BSComm application.
     * @throws InvalidParametersException Thrown when parameters are not valid.
     * @throws SendingMessageException Thrown when sending is not successful.
     * @throws TimeExpiredException Thrown when time of sending message expired.
     * @throws DeviceBusyException Thrown when mote is busy. There is a communication with basestation.
     */
    private void send() throws InvalidParametersException, ConnectException,
            SendingMessageException, TimeExpiredException, DeviceBusyException {

        // checking address
        if (isInvalid(getAddress())) {
            throw new InvalidParametersException();
        }

        try {
            Integer.valueOf(getAddress());
        } catch (NumberFormatException e) {
            throw new InvalidParametersException();
        }

        // creates new message according to protocol
        MessageConstructor messageConstructor = new MessageConstructor();
        String message = messageConstructor.constructAgentMessage(Integer.valueOf(address), getAgentBean());


        // gets port
        int port = DAOFactory.getDAOFactory().getSettingDAO().read().getPort();

        // creating new connector
        ClientSocket clientSocket = null;
        try {
            clientSocket = new ClientSocket(port);
        } catch (UnknownHostException ex) {
            throw new ConnectException();
        } catch (IOException ex) {
            throw new ConnectException();
        }


        // sending
        if (clientSocket != null) {
            try {
                clientSocket.send(message);
                clientSocket.close();
            } catch (IOException ex) {
                throw new SendingMessageException();
            } catch (ClassNotFoundException ex) {
                throw new SendingMessageException();
            } catch (CommunicationErrorException ex) {
                throw new SendingMessageException();
            }
        } else {
            throw new ConnectException();
        }

    }

    /**
     * Saves or update agent to database.
     * @throws DatabaseHandlingException This is thrown when the operation of saving or updating failed.
     */
    private void save() throws DatabaseHandlingException {

        if (aId > 0) {
            // getting existing
            Agent agent;
            try {
                agent = DAOFactory.getDAOFactory().getAgentDAO().read(aId);
            } catch (DatabaseErrorException ex) {
                throw new DatabaseHandlingException();
            }
            // setting new values
            agent.setAgentId(agentBean.getAgentId());
            agent.setClassId(agentBean.getClassId());
            agent.setDateSend(null);
            agent.setName(agentBean.getName());
            agent.setDescription(agentBean.getDescription());
            agent.setPlanBase(agentBean.getPlanBase());
            agent.setPlan(agentBean.getPlan());
            agent.setBeliefBase(agentBean.getBeliefBase());
            agent.setInputBase(agentBean.getInputBase());

            // update
            DAOFactory.getDAOFactory().getAgentDAO().update(agent);
        } else {

            agents.add(getAgentBean());

            try {
                // save it to DB
                DAOFactory.getDAOFactory().getAgentDAO().save(getAgentBean());
            } catch (Exception e) {
                throw new DatabaseHandlingException();
            }
        }
    }

    /**
     * Saves or update agent to database when is agent send do the WSN.
     * Saves time send. This is used in tracking agent.
     * @throws DatabaseHandlingException This is thrown when the operation of saving or updating failed.
     */
    private void savesend() throws DatabaseHandlingException {

        Date newDate = new Date();
        if (aId > 0) {
            // getting existing
            Agent agent;
            try {
                agent = DAOFactory.getDAOFactory().getAgentDAO().read(aId);
            } catch (DatabaseErrorException ex) {
                throw new DatabaseHandlingException();
            }
            // setting new values
            agent.setAgentId(agentBean.getAgentId());
            agent.setClassId(agentBean.getClassId());
            agent.setDateSend(newDate);
            agent.setName(agentBean.getName());
            agent.setDescription(agentBean.getDescription());
            agent.setPlanBase(agentBean.getPlanBase());
            agent.setPlan(agentBean.getPlan());
            agent.setBeliefBase(agentBean.getBeliefBase());
            agent.setInputBase(agentBean.getInputBase());

            // update
            DAOFactory.getDAOFactory().getAgentDAO().update(agent);
        } else {

            agentBean.setDateSend(newDate);
            agents.add(getAgentBean());

            try {
                // save it to DB
                DAOFactory.getDAOFactory().getAgentDAO().save(getAgentBean());
            } catch (Exception e) {
                throw new DatabaseHandlingException();
            }
        }
    }

    /**
     * Get the value of agents
     *
     * @return the value of agents
     */
    public List<Agent> getAgents() {
        return agents;
    }

    /**
     * Set the value of agents
     *
     * @param agents new value of agents
     */
    public void setAgents(List<Agent> agents) {
        this.agents = agents;
    }

    /**
     * Get the value of agentBean
     *
     * @return the value of agentBean
     */
    public Agent getAgentBean() {
        return agentBean;
    }

    /**
     * Set the value of agentBean
     *
     * @param agentBean new value of agentBean
     */
    public void setAgentBean(Agent agentBean) {
        this.agentBean = agentBean;
    }

    /**
     * Get the value of address
     *
     * @return the value of address
     */
    public String getAddress() {
        return address;
    }

    /**
     * Set the value of address
     *
     * @param address new value of address
     */
    public void setAddress(String address) {
        this.address = address;
    }

    /**
     * Get the value of param
     *
     * @return the value of param
     */
    public String getParam() {
        return param;
    }

    /**
     * Set the value of param
     *
     * @param param new value of param
     */
    public void setParam(String param) {
        this.param = param;
    }

    /**
     * Get the value of aId
     *
     * @return the value of aId
     */
    public long getAId() {
        return aId;
    }

    /**
     * Set the value of aId
     *
     * @param aId new value of aId
     */
    public void setAId(long aId) {
        this.aId = aId;
    }
}
