// -----------------------------------------------------------
// LOGParser.java
// -----------------------------------------------------------

package cz.vutbr.wsagent.util;

import cz.vutbr.wsagent.entity.Message;
import cz.vutbr.wsagent.entity.Agent;
import cz.vutbr.wsagent.db.hibernate.DAOFactory;
import cz.vutbr.wsagent.db.hibernate.DatabaseErrorException;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.Collections;

/**
 * This class is used for parsing logs message.
 * @author Marek Houšť
 */
public class LOGParser {

    private static final String LOG_PATTERN = "\\(LOG,[0-9 ]+,[0-9 ]+,[0-9 ]+,[0-9 ]+,[0-9 ]+\\)";
    private static final String LOG_DIVIDER = ",";
    private static final int LOG_FROM_NODE_FIELD = 1;
    private static final int FROM_FIELD = 2;
    private static final int AGENT_ID_FIELD = 3;
    private static final int AGENT_CLASS_FIELD = 4;
    private static final int COUNTER_FIELD = 5;
    private static final String END_OF_PATTERN = ("\\)");

    /**
     * This method finds out every message with LOG values from list of all given messages.
     * @param allRecords all given messages.
     * @return Returns list of messages with LOG values.
     */
    public List getCommonLogDates(List allRecords) {

        // pattern for matching LOG values
        Pattern pattern = Pattern.compile(LOG_PATTERN);

        // new list of records.
        List logsRecordList = new ArrayList<LOGRecord>();

        // date record
        LOGRecord logRecord = null;

        // go through all records
        // -----------------------------------------
        for (int i = 0; i < allRecords.size(); i++) {

            Message message = (Message) allRecords.get(i);

            if (message == null || message.getValue() == null) {
                continue;
            }

            // erase empty spaces
            String mess = message.getValue();
            message.setValue(mess.replaceAll(" ", ""));

            // check for pattern
            Matcher m = pattern.matcher(message.getValue());

            // any match ? - determines only LOG records
            if (m.find()) {

                if (message.isIsNew()) {

                    // creating new record
                    logRecord = new LOGRecord();

                    logRecord = fillLogRecord(logRecord, message);

                    // add new Record
                    logsRecordList.add(logRecord);
                    
                }
                
            }

        }
        
        return logsRecordList;
        
    }

    /**
     * This method fill log record from message.
     * @param logRecord log record witch is filling.
     * @param message message with log record.
     * @return Returns filled logRecord.
     */
    private LOGRecord fillLogRecord(LOGRecord logRecord, Message message) {

        Pattern pattern = Pattern.compile(LOG_PATTERN);
        Matcher m = pattern.matcher(message.getValue());

        // find the match
        if(m.find()) {

            // parse message
            String[] logValues = m.group().split(LOG_DIVIDER);
            // LOG_FROM_NODE_FIELD = 1
            // FROM_FIELD = 2
            // AGENT_ID_FIELD = 3
            // AGENT_CLASS_FIELD = 4
            // COUNTER_FIELD = 5
            String logFromNodeString = logValues[LOG_FROM_NODE_FIELD];
            String fromString = logValues[FROM_FIELD];
            String agentIdString = logValues[AGENT_ID_FIELD];
            String agentClassString = logValues[AGENT_CLASS_FIELD];
            String[] tempCounterString = logValues[COUNTER_FIELD].split(END_OF_PATTERN);
            String counterString = tempCounterString[0];
            
            // fill logRecord
            logRecord.setLogFromNode(Long.valueOf(logFromNodeString));
            logRecord.setFrom(Long.valueOf(fromString));
            logRecord.setAgentId(Long.valueOf(agentIdString));
            logRecord.setAgentClass(Long.valueOf(agentClassString));
            logRecord.setCounter(Long.valueOf(counterString));
            
        }

        return logRecord;
    }

    /**
     * This method return list of mote address.
     * @param logsRecords all parsed log records.
     * @return Returns list of mote address.
     */
    public List getMotesAddressFromLogs(List logsRecords) {

        List motesAddress = new ArrayList<Long>();
        long moteAddress;
        Long temp_moteAddress;
        boolean isSameMoteAddress = false;

        // go through all records
        // -----------------------------------------
        for (int i = 0; i < logsRecords.size(); i++) {

            LOGRecord logRecord = (LOGRecord) logsRecords.get(i);

            if (logRecord == null) {
                continue;
            }

            moteAddress = logRecord.getLogFromNode();
            
            for (int j = 0; j < motesAddress.size(); j++) {
                temp_moteAddress = (Long) motesAddress.get(j);
                if (moteAddress == temp_moteAddress) {
                    isSameMoteAddress = true;
                    break;
                }
            }
            if (!isSameMoteAddress) {
                motesAddress.add(moteAddress);
            }
            isSameMoteAddress = false;
        }

        return motesAddress;
    }

    /**
     * Method creates a single movement.
     * @param logRecord
     * @return trace
     */
    private Trace getTrace(LOGRecord logRecord) {

        Trace trace = new Trace();

        trace.setFrom(logRecord.getFrom());
        trace.setTo(logRecord.getLogFromNode());
        trace.setCounter(logRecord.getCounter());

        return trace;
    }

    /**
     * Method creates a categorized agent.
     * @param logRecord
     * @return tracingAgent
     */
    private TracingAgent getTracingAgent(LOGRecord logRecord) {
        
        TracingAgent tracingAgent = null;
        tracingAgent = new TracingAgent();
        
        tracingAgent.setAgentClass(logRecord.getAgentClass());
        tracingAgent.setAgentId(logRecord.getAgentId());
        
        List motesAddressWhereWas = new ArrayList<Long>();
        motesAddressWhereWas.add(logRecord.getFrom());
        motesAddressWhereWas.add(logRecord.getLogFromNode());
        tracingAgent.setMotesAddressWhereWas(motesAddressWhereWas);
        
        List traceList = new ArrayList<Trace>();
        traceList.add(getTrace(logRecord));
        tracingAgent.setTraceList(traceList);

        // get list of dates
        Agent temp_agent = null;
        boolean agentInDatabase = true;

        try {
            temp_agent = DAOFactory.getDAOFactory().getAgentDAO().read(logRecord.getAgentId(), logRecord.getAgentClass());
        } catch (DatabaseErrorException ex) {
            agentInDatabase = false;
        }
        if (agentInDatabase) {
            tracingAgent.setDateSend(temp_agent.getDateSend());
        } else {
            tracingAgent.setDateSend(null);
        }
        
        return tracingAgent;
    }

    /**
     * Method adds only those address that have not been added yet.
     * @param moteAddress
     * @param motesAddressWhereWas
     * @return motesAddressWhereWas
     */
    private List addOnlyNewMotesAddress(Long moteAddress, List motesAddressWhereWas) {

        Long temp_moteAddress;
        boolean isSameMoteAddress = false;
        
        for(int i = 0; i < motesAddressWhereWas.size(); i++) {
            temp_moteAddress = (Long) motesAddressWhereWas.get(i);
            if (temp_moteAddress == moteAddress) {
                isSameMoteAddress = true;
                break;
            }
        }
        if(!isSameMoteAddress) {
            motesAddressWhereWas.add(moteAddress);
        }

        return motesAddressWhereWas;
    }

    /**
     * This method finds out all agents and motes where was.
     * @param logsRecords all parsed log records.
     * @return Returns list of tracing agents with motes where was.
     */
    public List getTracingAgentList(List logsRecords) {

        // new list of tracingAgentList.
        List tracingAgentList = new ArrayList<TracingAgent>();
        
        /*
         * tracingAgent record
         * - agentClass
         * - agentId
         * - List<Long> motesAddressWhereWas
         * - List<Trace> traceList
         *
         * trace contain
         * - from
         * - to
         * - counter
         */
        TracingAgent tracingAgent = null;
        TracingAgent temp_tracingAgent = null;
        boolean isSameTracingAgent = false;

        // go through all records
        // -----------------------------------------
        for (int i = 0; i < logsRecords.size(); i++) {

            /*
             * logRecord
             * - date
             * - logFromNode (in trace has practical importance "to")
             * - from
             * - agentId
             * - agentClass
             * - counter
             */
            LOGRecord logRecord = (LOGRecord) logsRecords.get(i);

            if (logRecord == null) {
                continue;
            }

            /*
             * We are creating new agent always.
             */
            tracingAgent = new TracingAgent();
            tracingAgent = getTracingAgent(logRecord);
            /*
             * j will be used in the future to add trace to existing agent in list of agents
             */
            int j = 0;

            for (j = 0; j < tracingAgentList.size(); j++) {
                temp_tracingAgent = new TracingAgent();
                temp_tracingAgent = (TracingAgent) tracingAgentList.get(j);
                if ((tracingAgent.getAgentClass() == temp_tracingAgent.getAgentClass()) && (tracingAgent.getAgentId() == temp_tracingAgent.getAgentId())) {
                    isSameTracingAgent = true;
                    break;
                }
            }
            /*
             * if (isSameTracingAgent)
             * - If we found agent with same class and id in the list of agents.
             * - We are adding to this agent only other trace to the list of trace
             * - and new mote address where was.
             *
             * else
             * - We are adding new agent to the list of agents.
             * 
             */
            if (isSameTracingAgent) {
                List temp_traceList = new ArrayList<Trace>();
                temp_traceList = temp_tracingAgent.getTraceList();
                List traceList = new ArrayList<Trace>();
                traceList = tracingAgent.getTraceList();
                /*
                 * If there is the agent with the same class and id.
                 * - traceList currently created agent will always have only one item
                 */
                temp_traceList.add(traceList.get(0));

                // sort each tracingAgent - his traceList by counter in class Trace
                Collections.sort(temp_traceList);

                List temp_motesAddressWhereWas = new ArrayList<Long>();
                temp_motesAddressWhereWas = temp_tracingAgent.getMotesAddressWhereWas();
                List motesAddressWhereWas = new ArrayList<Long>();
                motesAddressWhereWas = tracingAgent.getMotesAddressWhereWas();
                /*
                 * If there is the agent with the same class and id.
                 * - motesAddressWhereWas currently created agent will always have only two item
                 */
                temp_motesAddressWhereWas = addOnlyNewMotesAddress((Long) motesAddressWhereWas.get(0), temp_motesAddressWhereWas);
                temp_motesAddressWhereWas = addOnlyNewMotesAddress((Long) motesAddressWhereWas.get(1), temp_motesAddressWhereWas);

                
                temp_tracingAgent.setMotesAddressWhereWas(temp_motesAddressWhereWas);
                temp_tracingAgent.setTraceList(temp_traceList);
                
                tracingAgentList.set(j, temp_tracingAgent);
                
            } else {

                tracingAgentList.add(tracingAgent);
                
            }
            isSameTracingAgent = false;
        }
        
        return tracingAgentList;
    }

}
