// -----------------------------------------------------------
// RSSIParser.java
// -----------------------------------------------------------
package cz.vutbr.wsagent.util;

import cz.vutbr.wsagent.entity.Message;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class is used for parsing all messages that contain some kind of data
 * values.
 * @author Martin Gábor
 */
public class RSSIParser {

    private static final String RSSI_PATTERN = "\\(NB,[0-9 ]+,[0-9 ]+\\)";
    private static final String RSSI_PATTERN_FOR_TRACES = "\\(M,[0-9 ]+,[0-9 ]+,[0-9 ]+\\)";
    private static final int SOURCE_ADDRESS_FIELD_FOR_TRACES = 1;
    private static final int DEST_ADDRESS_FIELD_FOR_TRACES = 2;
    private static final int SIGNAL_STRENGTH_FIELD_FOR_TRACES = 3;

    private static final String RSSI_DIVIDER = ",";
    private static final int DEST_ADDRESS_FIELD = 1;
    private static final int SIGNAL_STRENGTH_FIELD = 2;
    private static final String END_OF_PATTERN = ("\\)");
    private static final int MAX_RSSI_VALUE = 255;
    private static final int MILI_SECONDS_INTERVAL = 2000; // 1 sec

    /***
     * This method parses message looking for patterns of rssi values.
     * It returns a list of RSSIObjects found in message.
     * @param message Input message, where probably are some rssi values.
     * @return Returns List of RSSIObjects.
     */
    public List<RSSIObject> getRSSI(Message message) {

        List rssiObjectList = new ArrayList<RSSIObject>();

        Pattern pattern = Pattern.compile(RSSI_PATTERN);
        Matcher m = pattern.matcher(message.getValue());

        // finding the matches
        while (m.find()) {

            // split values
            String[] values = m.group().split(RSSI_DIVIDER);
            String destAddress = values[DEST_ADDRESS_FIELD];
            String[] rssi = values[SIGNAL_STRENGTH_FIELD].split(END_OF_PATTERN);

            // else

            // create new instance of RSSI Object
            RSSIObject rssiObject = new RSSIObject();

            // set the values
            rssiObject.setSourceAddress(message.getMote().getAddress());
            rssiObject.setDestAddress(destAddress);
            rssiObject.setStrength(getPercentageRSSI(Integer.valueOf(rssi[0])));

            // add object to Object List
            rssiObjectList.add(rssiObject);
        }


        // return list of RSSIObject
        return rssiObjectList;
    }

    /***
     * This method parses message looking for patterns of rssi values.
     * It returns a list of RSSIObjects found in message.
     * @param message Input message, where probably are some rssi values.
     * @return Returns List of RSSIObjects.
     */
    public List<RSSIObject> getRSSIForTraces(Message message) {

        List rssiObjectList = new ArrayList<RSSIObject>();

        Pattern pattern = Pattern.compile(RSSI_PATTERN_FOR_TRACES);
        Matcher m = pattern.matcher(message.getValue());

        // finding the matches
        while (m.find()) {

            // split values
            String[] values = m.group().split(RSSI_DIVIDER);
            // SOURCE_ADDRESS_FIELD_FOR_TRACES = 1;
            // DEST_ADDRESS_FIELD_FOR_TRACES = 2;
            // SIGNAL_STRENGTH_FIELD_FOR_TRACES = 3;
            String sourceAddress = values[SOURCE_ADDRESS_FIELD_FOR_TRACES];
            String destAddress = values[DEST_ADDRESS_FIELD_FOR_TRACES];
            String[] rssi = values[SIGNAL_STRENGTH_FIELD_FOR_TRACES].split(END_OF_PATTERN);

            // else

            // create new instance of RSSI Object
            RSSIObject rssiObject = new RSSIObject();

            // set the values
            rssiObject.setSourceAddress(sourceAddress);
            rssiObject.setDestAddress(destAddress);
            rssiObject.setStrength(getPercentageRSSI(Integer.valueOf(rssi[0])));

            // add object to Object List
            rssiObjectList.add(rssiObject);
        }


        // return list of RSSIObject
        return rssiObjectList;
    }

    /**
     * Returns percentage value of RSSI
     * @param givenValue value that we want to convert to percentage value.
     * @return Returns percentage value of given value.
     */
    public int getPercentageRSSI(int givenValue) {

        return (100 * givenValue) / MAX_RSSI_VALUE;
    }

    /**
     * This method finds out every message with RSSI values from list of all given messages.
     * @param allRecords all given messages.
     * @return Returns list of messages with RSSI values.
     */
    public List getCommonRssiDates(List allRecords) {

        // pattern for matching RSSI values
        Pattern pattern = Pattern.compile(RSSI_PATTERN);

        // new list of records.
        List rssiRecordList = new ArrayList<RSSIRecord>();

        // declaring last values
        Date lastDate = new Date(0);
        int lastId = 0;


        // list of objects that will be connected to concrete record
        // which is described by the date of discovery.
        List rssiObjectList = null;
        // date record
        RSSIRecord record = null;


        // go through all records
        // -----------------------------------------
        for (int i = 0; i < allRecords.size(); i++) {

            Message message = (Message) allRecords.get(i);

            if (message == null || message.getValue() == null) {
                continue;
            }

                    // erase empty spaces
            String mess = message.getValue();
            message.setValue(mess.replaceAll(" ", ""));

            // check for pattern
            Matcher m = pattern.matcher(message.getValue());


            // any match ? - determines only RSSI records
            if (m.find()) {

                /*
                 * Splitting up the lists of RSSI object to records with concrete date.
                 * WE use MILI_SECONDS_INTERVAL to define the error interval of coming RSSI agent messages.
                 * The interval defines that every RSSI record belonging to particular discovery
                 * is putting to the same place in the list.
                 */
                if (message.getDate().getTime() > lastDate.getTime() + (MILI_SECONDS_INTERVAL * 2)) {

                    // creating new record - the date of message was not in the interval.
                    record = new RSSIRecord();
                    record.setDate(message.getDate());
                    record.setId(lastId);

                    // create new list of object
                    rssiObjectList = getRSSI(message);
                    record.setRssiObjectList(rssiObjectList);

                    // add new Record
                    rssiRecordList.add(record);

                    // indrement last values
                    lastId++;
                    lastDate = message.getDate();

                    // the date of message is within the interval.
                } else {

                    //new objects
                    List newObjects = getRSSI(message);

                    // check occurence
                    rssiObjectList = checkOccurenceInOldList(newObjects, rssiObjectList);

                    // save all objects
                    record.setRssiObjectList(rssiObjectList);
                }
            }
        }
        // -----------------------------------------


        // return the final list of records where every record contain date describing
        // itself and list of objects containing values with source and dest. Address and
        // the rssi.
        return rssiRecordList;

    }

    /**
     * Practically same as getCommonRssiDates. Only for purpose of class WSTracingAgents.
     *
     * This method finds out every message with RSSI values from list of all given messages.
     * @param allRecords all given messages.
     * @return Returns record with RSSI values.
     */
    public RSSIRecord getCommonRssiDatesForTraces(List allRecords) {

        // pattern for matching RSSI values
        Pattern pattern = Pattern.compile(RSSI_PATTERN_FOR_TRACES);

        int lastId = -1;

        // list of objects that will be connected to concrete record
        // which is described by the date of discovery.
        List rssiObjectList = null;
        List newObjects = null;
        // date record
        RSSIRecord record = new RSSIRecord();

        record.setId(lastId);
        lastId = 0;

        // go through all records
        // -----------------------------------------
        for (int i = 0; i < allRecords.size(); i++) {

            Message message = (Message) allRecords.get(i);

            if (message == null || message.getValue() == null) {
                continue;
            }

            if (message.isIsNew()) {

                // erase empty spaces
                String mess = message.getValue();
                message.setValue(mess.replaceAll(" ", ""));

                // check for pattern
                Matcher m = pattern.matcher(message.getValue());

                // any match ? - determines only RSSI records
                if (m.find()) {

                    record.setDate(message.getDate());
                    record.setId(lastId);

                    if (lastId == 0) {
                        rssiObjectList = getRSSIForTraces(message);
                        newObjects = rssiObjectList;
                    } else {
                        newObjects = getRSSIForTraces(message);
                    }

                    // check occurence
                    rssiObjectList = checkOccurenceInOldList(newObjects, rssiObjectList);

                    record.setRssiObjectList(rssiObjectList);

                    // increment last values
                    lastId++;

                }
            }
        }
        // -----------------------------------------

        return record;
    }

    /**
     * This method compares two lists and finds the same nodes connections.
     * For example: (6,8) = (8,6).
     * It searches in old list of RSSI objects and looks for the same node connections in new list
     * of RSSI objects.
     * If it is found makes arithmetic mean of strength and save it to the object in old list.
     * The method returns old list of RSSI objects with added objects from
     * new list that have not been found in old list.
     * @param newObjectList The list of new RSSI objects.
     * @param oldObjectList The list of old RSSI objects.
     * @return Returns old list of RSSI objects with added objects from
     * new list that have not been found in old list.
     */
    private List checkOccurenceInOldList(List newObjectList, List oldObjectList) {

        // check if the objects is already in the list
        // ----------------------------------------
        RSSIObject newObject;
        RSSIObject oldObject;
        List newObjectsToAdd = new ArrayList<RSSIObject>();

        for (Iterator newIter = newObjectList.iterator(); newIter.hasNext();) {

            int i = 0;
            newObject = (RSSIObject) newIter.next();
            // is it there ?
            boolean isFound = false;
            
            for (Iterator oldIter = oldObjectList.iterator(); oldIter.hasNext();i++) {

                oldObject = (RSSIObject) oldIter.next();

                // check if sourceAddr1 == destAddr2 and sourceAddrw == destAddrq.
                if (((newObject.getSourceAddress().compareTo(oldObject.getDestAddress()) == 0) &&
                      (newObject.getDestAddress().compareTo(oldObject.getSourceAddress()) == 0)) ||

                        ((newObject.getSourceAddress().compareTo(oldObject.getSourceAddress()) == 0) &&
                        (newObject.getDestAddress().compareTo(oldObject.getDestAddress()) == 0))


                        ) {

                    // create arithmetic mean of strengths
                    int newStrength = (newObject.getStrength() + oldObject.getStrength()) / 2;

                    // set new strength
                    oldObject.setStrength(newStrength);
                    oldObjectList.set(i,oldObject);

                    isFound = true;
                    continue;
                }
            }
                            if(!isFound)
                    newObjectsToAdd.add(newObject);
        }
        // ----------------------------------------

        // add new objects
        oldObjectList.addAll(newObjectsToAdd);

        return oldObjectList;

    }

}
