// -----------------------------------------------------------
// TopologyDrawer.java
// -----------------------------------------------------------

package cz.vutbr.wsagent.util;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.List;
import javax.imageio.ImageIO;

/**
 * The class for drawing topology to
 * @author Martin Gábor
 */
public class TopologyDrawer {

    /* COLOR PROPERTIES */
    // backgroundColor
    private static final Color bgColor = new Color(0xFB,0xF7,0xF7);
    // color of text beside nodes 6c8395
    private static final Color percentColor = new Color(0x6C,0x83,0x95);
    // color of text beside nodes 6c8395
    private static final Color nodeTextColor = new Color(0x77,0x00,0x00);
    // color of lines
    private static final Color linesColor = new Color(0x87,0x6f,0x6f);

    private BufferedImage bufferedImage;
    private String webPath;


    /**
     * Constructor for creating new instance. It prepares image to draw in it.
     * @param width The width of scene.
     * @param height The heightof scene.
     * @param inputPath The path where should be all images saved.
     */
    public TopologyDrawer(int width, int height, String inputPath){

        // set the path of images 
        webPath = inputPath;

        // create new buffered image
        bufferedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
        
        // creating graphics
        Graphics g = bufferedImage.getGraphics();
        Graphics2D g2d = (Graphics2D) g;

        // formatting graphic properties
        g2d.setBackground(bgColor);
        g2d.clearRect(0, 0, width, height);

        g2d.setColor(bgColor);
        g2d.fillRect(0, 0, width, height);

    }

    /**
     * This is used for drawing topology image.
     * @param moteList The list of all objects - points and lines.
     * @param border Window border.
     */
    public void drawTopology(List<GeometryObject> moteList, RectangleObject border) {

        int size = moteList.size();

        for(int i = 0; i < size; i++){
            GeometryObject geoObject = (GeometryObject) moteList.get(i);

            // if it's point object - it is mote
            if(geoObject.getName().compareTo(GeometryObject.POINTNAME) == 0){
                try {
                    PointObject point = (PointObject) geoObject;
                    // draw mote
                    point.setX(point.getX()-border.getX1());
                    point.setY(point.getY()-border.getY1());
                    drawMote(point);
                } catch (IOException ex) {
                }
                
            }
            // if it is line
            else if(geoObject.getName().compareTo(GeometryObject.LINENAME) == 0){
                try {
                    LineObject line = (LineObject) geoObject;
                    // draw line
                    line.setX1(line.getX1()-border.getX1());
                    line.setY1(line.getY1()-border.getY1());
                    line.setX2(line.getX2()-border.getX1());
                    line.setY2(line.getY2()-border.getY1());
                    drawLine(line);
                } catch (IOException ex) {
                }
            }
        }

    }

    /**
     * This method is for drawing topology image.
     * @param pathFile the file.
     * @exception IOException Thrown when the given path is incorrect.
     */
    public void generateTopologyImage(String pathFile) throws IOException {

        // creating graphics
        Graphics g = bufferedImage.getGraphics();
        Graphics2D g2d = (Graphics2D) g;

        g2d.dispose();

        // writes the image to the disk of given path
        ImageIO.write(bufferedImage, "jpg", new File(webPath+pathFile));
    }

    /**
     * Method draws mote. Coordinates are the centre of object.
     * @param point The point object.
     */
    private void drawMote(PointObject point) throws IOException{

        Graphics g = bufferedImage.getGraphics();
        Graphics2D g2d = (Graphics2D) g;

        // mote drawing
        String moteSourceFile = webPath+"/ground/images/topology/mote.jpg";
        File moteFile = new File(moteSourceFile);
        Image buffImage = ImageIO.read(moteFile);

        g2d.drawImage(buffImage, point.getX()-30, point.getY()-16, Color.red, null);

        int length;
        // text mote
        g2d.setFont(new Font("Goudy Handtooled BT", Font.BOLD, 12));
        g2d.setPaint(nodeTextColor);
        length = (point.getId()).length();
        g2d.drawString(point.getId(), point.getX()-(5*length), point.getY()+20);

        // text mote
        g2d.setFont(new Font("Goudy Handtooled BT", Font.PLAIN, 12));
        g2d.setPaint(nodeTextColor);

        length = (""+point.getX()+""+point.getY()).length();
        g2d.drawString("["+point.getX()+","+point.getY()+"]", point.getX()-(5*length), point.getY()+35);
    }

    /**
     * Method draws line with given points.
     * @param line The line object.
     */
    private void drawLine(LineObject line) throws IOException{

        Graphics g = bufferedImage.getGraphics();
        Graphics2D g2d = (Graphics2D) g;

        // line
        float dash[] = {10.0f};
        g2d.setStroke(new BasicStroke(1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, dash, 0.0f));
        g2d.setColor(linesColor);
        g2d.drawLine(line.getX1(), line.getY1(), line.getX2(), line.getY2());

        // middle point
        int midX = line.getX1()+((line.getX2()-line.getX1())/2);
        int midY = line.getY1()+((line.getY2()-line.getY1())/2);

        // text strength
        g2d.setFont(new Font("Goudy Handtooled BT", Font.ITALIC, 11));
        g2d.setPaint(percentColor);
        g2d.drawString(""+line.getStrength()+"%", midX+10, midY+10);

    }

    /**
     * Get the value of webPath
     *
     * @return the value of webPath
     */
    public String getWebPath() {
        return webPath;
    }

    /**
     * Set the value of webPath
     *
     * @param webPath new value of webPath
     */
    public void setWebPath(String webPath) {
        this.webPath = webPath;
    }


    /**
     * Get the value of bufferedImage
     *
     * @return the value of bufferedImage
     */
    public BufferedImage getBufferedImage() {
        return bufferedImage;
    }

    /**
     * Set the value of bufferedImage
     *
     * @param bufferedImage new value of bufferedImage
     */
    public void setBufferedImage(BufferedImage bufferedImage) {
        this.bufferedImage = bufferedImage;
    }
}
