// -----------------------------------------------------------
// TopologyMath.java
// -----------------------------------------------------------

package cz.vutbr.wsagent.util;

import java.util.ArrayList;
import java.util.List;

/**
 * This class is for generating points and lines and all required objects for topology.
 * @author Martin Gábor
 */
public class TopologyMath {

    // distance between 2 objects with the smallest RSSI
    private static final int MAXDISTANCE = 600;
    // first point
    private static final int FIRSTX = 5000;
    private static final int FIRSTY = 5000;
    private static final int MINXBORDER = 0;
    private static final int MINYBORDER = 0;
    private static final int MAXXBORDER = 10000;
    private static final int MAXYBORDER = 10000;
    private static final int PADDING = 50;
    // number of steps used when searching unknown distance
    private static final int STEP_COUNT = 25;
    // unknown distance step - it is used to determine length of two object which do not
    // see each other
    private static final int UNKNOWN_DIST_STEP = 15;
    // known distance step - it is used to determine length of two object which see each other
    // but the received strength is slower than it should be
    private static final int KNOWN_DIST_STEP = 5;


    /**
     * The main method of determining the final points of toplogy.
     * Iterates through records of rssiObjectList containing the pair of addresses and it's
     * strength.
     * @param rssiObjectList The list of <code>RSSIObject</code> instances.
     * @return Returns the list of all points and lines of the topology.
     */
    public List<GeometryObject> getTopologyPoints(List<RSSIObject> rssiObjectList)
            throws InvalidInputStrengthsException {


        List resultPoints = new ArrayList<GeometryObject>();
        // copy to new list
        ArrayList<RSSIObject> helpList = (ArrayList<RSSIObject>) ((ArrayList) rssiObjectList).clone();

        // if the source is empty return empty result
        if (helpList.isEmpty()) {
            return resultPoints;
        }


        // ------------
        // pick first
        RSSIObject rssiObject = (RSSIObject) helpList.get(0);
        PointObject firstObject = new PointObject(FIRSTX, FIRSTY, rssiObject.getSourceAddress());
        resultPoints.add(firstObject);

        // pick second
        int distance = evaluateDistanceFromStrength(rssiObject.getStrength());
        PointObject secondObject = new PointObject(FIRSTX, FIRSTY + distance, rssiObject.getDestAddress());
        resultPoints.add(secondObject);

        // delete first and second
        helpList.remove(0);
        // ------------


        RSSIObject modifRSSIObj = new RSSIObject();
        while (modifRSSIObj != null) {

            modifRSSIObj = getExistingNode(helpList, resultPoints);


            // if found something
            if (modifRSSIObj != null) {

                PointObject finalObject = getPointPosition(helpList, modifRSSIObj, resultPoints);

                if (finalObject == null) {
                    throw new InvalidInputStrengthsException();
                }

                finalObject.setId(modifRSSIObj.getDestAddress());
                resultPoints.add(finalObject);

            }
        }
        // ----------------------------------------------------------

        // get lines
        List lines = getLines(rssiObjectList, resultPoints);

        // add lines to result
        resultPoints.addAll(lines);



        return resultPoints;
    }

    /**
     * This method returns an object with determined position. It is used when we recommend to
     * calculate a position of object that we want add to the topology depending on existing points of topology.
     * @param rssiObjectList The list of <code>RSSIObject</code> records.
     * @param rssiObject The object containing the pair of address where the first address is an address
     * of existing topology point and the second is an address of object that we want to find.
     * @param resultPoints The list of existing topology points.
     * @return Returns found point.
     * @throws InvalidInputStrengthsException This is thrown when we cannot determine the position of given
     * values of rssiObjectList.
     */
    private PointObject getPointPosition(List<RSSIObject> rssiObjectList, RSSIObject rssiObject,
            List<GeometryObject> resultPoints) throws InvalidInputStrengthsException {


        // the final position
        PointObject resultObject = null;


        // Circles made from known distances - motes that see each other with
        // specific strength
        List knownCircles = new ArrayList<CircleObject>();

        // Circles made from unknown distances. The radius is >= MAXDISTANCE.
        List unknownCircles = new ArrayList<CircleObject>();



        int resultSize = resultPoints.size();
        // iterate the list of motes with determined position
        // looking for all distances
        // --------------------------------------------------
        for (int i = 0; i < resultSize; i++) {

            // get objects
            GeometryObject geomObject = resultPoints.get(i);
            // if it is not Point continue
            if (geomObject.getName().compareTo(GeometryObject.POINTNAME) != 0) {
                continue;
            }
            // retype to point
            PointObject point = (PointObject) geomObject;


            // new circle
            CircleObject circle;

            // if it is the point we picked from rssiObjectList
            if (point.getId().compareTo(rssiObject.getSourceAddress()) == 0) {

                // get distance
                int distance = evaluateDistanceFromStrength(rssiObject.getStrength());

                // create new instance of circle
                circle = new CircleObject(point.getX(), point.getY(),
                        distance);

                // known point for sure
                knownCircles.add(circle);

                // find other points
                // if the record is not found
            } else {

                // find out the pair from list of RSSI Objects and get their strength.
                // if the pair is not found return 0.
                int strength = getStrengthFromList(point.getId(),
                        rssiObject.getDestAddress(), rssiObjectList);

                // get distance
                int distance = evaluateDistanceFromStrength(strength);

                // create new instance of circle
                circle = new CircleObject(point.getX(), point.getY(), distance);

                // if it is unknow distance
                if (distance == MAXDISTANCE) {
                    unknownCircles.add(circle);
                } // known distance
                else {
                    knownCircles.add(circle);
                }

            }
        }
        // -------------------------------------------------

        resultObject = checkAllPossibilities(knownCircles, unknownCircles, resultPoints);


        return resultObject;
    }


    /**
     * This method returns the most distant object of <code>testPoints</code>. The distance is measured
     * with each of <code>resultPoints</code> objects.
     * @param testPoints The list points where we finding the most distant.
     * @param resultPoints The list of existing object in our topology.
     * @throws InvalidInputStrengthsException Thrown when is not found any point.
     */
    private PointObject measureTheMostDistant(List<PointObject> testPoints, List<GeometryObject> resultPoints)
            throws InvalidInputStrengthsException {

        // the actual longest distance
        int currentDistance = 0;

        // retype to point
        PointObject distantPoint = null;

        int testSize = testPoints.size();
        int resultSize = resultPoints.size();

        for (int j = 0; j < testSize; j++) {

            PointObject testPoint = testPoints.get(j);

            // iterate the list of motes with determined position
            // looking for all distances
            // --------------------------------------------------
            int distance = 0;
            for (int i = 0; i < resultSize; i++) {

                // get objects
                GeometryObject geomObject = resultPoints.get(i);
                // if it is not Point continue
                if (geomObject.getName().compareTo(GeometryObject.POINTNAME) != 0) {
                    continue;
                }
                // retype to point
                PointObject point = (PointObject) geomObject;

                distance += testPoint.determineDistance(point);
            }

            // if further
            if (distance > currentDistance) {

                // save the point
                distantPoint = testPoint;
                // new current
                currentDistance = distance;
            }

        }

        if (distantPoint == null) {
            throw new InvalidInputStrengthsException();
        }

        return distantPoint;
    }


    /**
     * This is used for finding the nearest intersections.
     * @param intersects all intersections
     * @return Returns the nearest intersection. The size of returned list is is all / 2.
     */
    private List findNearestIntersects(List<PointObject>intersects){

        long MAXVALUE = Long.MAX_VALUE;

        int allSize = intersects.size();
        int arrSize = allSize;
        // get the number of circles
        int number = allSize/2;

        // the array of distances
        long[] distances = new long[allSize];

        // get all distances
        for(int i = 0; i < allSize; i++){

            long distance = 0;
            for(int j = 0; j < allSize; j++){
                distance+= intersects.get(i).determineDistance(intersects.get(j));
            }
            distances[i] = distance;
        }

        List result = new ArrayList<PointObject>();

        int i = 0;
        int minDistIndex = 0;
        long minDistance = distances[0];

        // we want only the nearest distances.
        while(arrSize > number){

            if(distances[i] < minDistance){
                minDistIndex = i;
                minDistance = distances[i];
            }

            i++;
            if(i >= allSize){
                result.add(intersects.get(minDistIndex));
                i = 0;
                distances[minDistIndex] = MAXVALUE;
                minDistance = MAXVALUE;
                arrSize--;
            }
        }

        return result;

    }

    /**
     * In process of determining the position of new points there are three variants depending on
     * how many points is the finding point connected with.
     * If there are three or two or one connection. These is chosen by the number of <code>known</code> circles.
     * Each of variants determines the poistion of finding point.
     * In case of two and one connection there are also taken unknown <code>circles</code> to measure the most
     * distant point from these circles.
     * @param known Circles with known radius.
     * @param unknown Circles with unknown radius - not connected with finding point.
     * @param resultPoints The list of already determind points of topology.
     * @return Returns the final object.
     * @throws InvalidInputStrengthsException Thrown when some intersections from given data cannot be found.
     */
    private PointObject checkAllPossibilities(List<CircleObject> known, List<CircleObject> unknown,
            List<GeometryObject> resultPoints) throws InvalidInputStrengthsException {

        List possibilities = new ArrayList<PointObject>();

        // how much circles do we have ???
        int knownSize = known.size();
        int unknownSize = unknown.size();


        // if we have 1 known and N unknown
        // -----------------------------------------------
        if (knownSize == 1) {

            // get the only circle in the known list
            CircleObject knownCircle = (CircleObject) known.get(0);


            // iterate through all unknown
            for (int i = 0; i < unknownSize; i++) {
                CircleObject unknownCircle = unknown.get(i);

                unknownCircle.decreaseRadius(UNKNOWN_DIST_STEP);

                List intersects = getUnknownIntersections(knownCircle, unknownCircle);

                // add all intersects
                possibilities.addAll(intersects);
            }


            // we have 2 known and N unknown
            // -------------------------------------------
        } else if (knownSize == 2) {

            List intersects = new ArrayList<PointObject>();

            // find 2 intersections
            intersects = getRelativeIntersection(known.get(0),
                    known.get(1));

            if (intersects.size() == 2) {
                possibilities.addAll(intersects);
            }

            // check if it is empty == error
            // this should not happen
            if (intersects.isEmpty()) {
                throw new InvalidInputStrengthsException();
            }


            // we have at least 3 known
            // ----------------------------------------------
        } else if (knownSize >= 3) {

            List intersects = new ArrayList<PointObject>();

            for (int i = 0; i < knownSize - 1; i++) {
                for (int j = i + 1; j < knownSize; j++) {

                    // get intersections of 2 and 3 circles
                    intersects.addAll(getRelativeIntersection((CircleObject) known.get(i),
                            (CircleObject) known.get(j)));
                }
            }

            //check if enough intersections
            if (intersects.size() <= 4) {
                throw new InvalidInputStrengthsException();
            }

            List nearestInterSects = findNearestIntersects(intersects);

            // build the polygon object from intersections
            PolygonObject polygon = new PolygonObject(nearestInterSects);

            // return middle Point of all intersections
            return polygon.getMiddlePoint();

        }
        // ------------------------------------------------------------


        // if known == 1 or known == 2
        // find the distant intersect
        PointObject distantObject = measureTheMostDistant(possibilities, resultPoints);


        return distantObject;
    }

    /**
     * This method is used when we are not absolutely sure if there are any intersections of
     * two cricles. It is looking for them in steps. In each step increases the radius of unknown
     * circle. The increased value is defined by constant UNKNOWN_DIST_STEP.
     * All found intersections are returned. If nothing found, empty list is returned.
     * @param known Circle with known radius.
     * @param unknown Circle with unknown radius.
     * @return Returns all intersections of two circles.
     */
    private List<PointObject> getUnknownIntersections(CircleObject known, CircleObject unknown){

        int step = 0;

        List intersects = new ArrayList<PointObject>();

        while (step < STEP_COUNT) {
            // get intersections
            List<PointObject> points = known.getIntersection(unknown);

            intersects.addAll(points);

            unknown.increaseRadius(UNKNOWN_DIST_STEP);
            step++;
        }

        return intersects;
    }


    /**
     * This returns intersections of two circles. It is determined in steps. The first
     * step that gives two intersections is final. In each step smaller radius of circles
     * is increased by value of constant KNOWN_DIST_STEP.
     * If nothing found, empty list is returned.
     * @param circle1 First circle.
     * @param circle2 Second circle.
     * @return Returns list of intersections.
     */
    private List<PointObject> getRelativeIntersection(CircleObject circle1, CircleObject circle2) {

        int step = 0;

        List intersects = new ArrayList<PointObject>();

        while (intersects.size() != 2 && step < STEP_COUNT) {
            // get intersections
            intersects = (circle1.getIntersection(circle2));

            //increase
            if (circle1.getR() > circle2.getR()) {
                circle2.increaseRadius(KNOWN_DIST_STEP);
            } else {
                circle1.increaseRadius(KNOWN_DIST_STEP);
            }
            step++;
        }

        return intersects;
    }

    /**
     * Returns true if params first couple of strings are equal to the second.
     * @param source1 Source string of first object.
     * @param dest1 Destination string of first object.
     * @param source2 Source string of second object.
     * @param dest2 Destination string of second object.
     * @return Returns true if couples of strings are equal.
     */
    private boolean isTheSameCouple(String source1, String dest1, String source2, String dest2) {

        if (source1.compareTo(source2) == 0 && dest1.compareTo(dest2) == 0) {
            return true;
        }

        if (source1.compareTo(dest2) == 0 && dest1.compareTo(source2) == 0) {
            return true;
        }

        return false;
    }


    /**
     * Searches the value of strength defined by source and destination address in the rssiObject list.
     * If there is no record of given addresses, zero strength is returned.
     * @param sourceAdrr Source address.
     * @param destAdrr Destination address.
     * @param rssiObjects The list of rssiObject records.
     * @see RSSIObject.
     */
    private int getStrengthFromList(String sourceAdrr, String destAdrr, List rssiObjects) {

        int size = rssiObjects.size();
        // searching
        for (int i = 0; i < size; i++) {

            RSSIObject objectIterator = (RSSIObject) rssiObjects.get(i);

            if (isTheSameCouple(sourceAdrr, destAdrr, objectIterator.getSourceAddress(),
                    objectIterator.getDestAddress())) {
                return objectIterator.getStrength();
            }

        }
        return 0;
    }

    /**
     * Searches the object of PointObject with specified id in the list of PointObject.
     * Returns true if the object is found.
     * @param id Id of object which we want to find.
     * @param resultPoints the list of PointObject.
     * @return Returns 
     */
    private boolean isInResultPoints(String id, List<PointObject> resultPoints) {

        int size = resultPoints.size();
        for (int i = 0; i < size; i++) {

            PointObject point = resultPoints.get(i);

            if (point.getId().compareTo(id) == 0) {
                return true;
            }

        }

        return false;
    }

    /**
     * Returns the number of connections of resultPoints and object specified by address.
     * The connection means that there is a record in list of objects which declares the pair of
     * addresses and it's strength.
     * @param address The address of object.
     * @param objects List of RSSIObject records.
     * @param resultPoints The list of already determined topology points.
     * @return Returns the number of connections.
     */
    private int getConnectionNumber(String address, List<RSSIObject> objects, List<PointObject> resultPoints) {

        int sizeResult = resultPoints.size();
        int sizeObjects = objects.size();
        int number = 0;

        for (int i = 0; i < sizeResult; i++) {
            for (int j = 0; j < sizeObjects; j++) {

                // check if it is the same couple
                if (isTheSameCouple((resultPoints.get(i)).getId(), address,
                        (objects.get(j)).getSourceAddress(), (objects.get(j)).getDestAddress())) // yes
                {
                    number++;
                }
            }
        }

        return number;
    }

    /**
     * Returns the number of <code>RSSIObject</code> records which sourceAddress or DestAddress is
     * <code>newAddress</code>.
     * @param newAddress Address that is searched in <code>RSSIObject</code> List.
     * @param objects List of <code>RSSIObjects</code>.
     * @return Returns number of <code>RSSIObjects</code> containg given address.
     */
    private int getNumberRecords(String newAddress, List<RSSIObject> objects) {

        int sizeObjects = objects.size();
        int number = 0;

        // iterate
        for (int j = 0; j < sizeObjects; j++) {

            // check if it is the same couple
            if ((objects.get(j).getSourceAddress().compareTo(newAddress)) == 0) {
                number++;
            } else if ((objects.get(j).getDestAddress().compareTo(newAddress)) == 0) {
                number++;
            }
        }

        return number;
    }

    /**
     * It returns new <code>RSSIObject</code>, where one address is and the second has not been determined yet.
     * It depends on amount of existed connections of searched mote with existing.
     * The address with highest connections is the winner. It is returned with address of existing mote.
     * The form is always (determined, winner, strength).
     * @param objects RSSIObject records with couples of addresses and their mutual strength.
     * @param resultPoints The List of points with determined positions.
     * @return Returns found RSSIObject(determined Addr, winner Addr, strength)
     */
    private RSSIObject getExistingNode(List<RSSIObject> objects, List<PointObject> resultPoints) {

        // looking for results
        boolean firstFound = false;
        boolean secondFound = false;
        boolean isResultSource = false;

        // final object
        RSSIObject finalObject = null;

        // size of RSSIObject List
        int sizeObject = objects.size();

        // amounts of connections
        int number;
        int bestNumber = -1;
        int bestSameNumber = -1;
        String newAddress;

        //
        for (int j = 0; j < sizeObject; j++) {

            RSSIObject object = objects.get(j);

            firstFound = isInResultPoints(object.getSourceAddress(), resultPoints);
            secondFound = isInResultPoints(object.getDestAddress(), resultPoints);
            // if the are both, skip it
            if (firstFound && secondFound) {
                continue;
            }

            // first is address of existing mote
            if (firstFound) {
                number = getConnectionNumber(object.getDestAddress(), objects, resultPoints);
                newAddress = object.getDestAddress();
                // second is address of existing mote
            } else if (secondFound) {
                number = getConnectionNumber(object.getSourceAddress(), objects, resultPoints);
                newAddress = object.getSourceAddress();
            } else {
                continue;
            }

            // looking for best number
            if (number >= bestNumber) {

                // found it
                // if there is the number of connection ->
                // find node with higher RSSIObject records
                if (number == bestNumber) {
                    // has it more RSSIObjects?
                    int numberAll = getNumberRecords(newAddress, objects);

                    // checking if more records
                    if (numberAll < bestSameNumber) {
                        continue;
                    } else {
                        bestSameNumber = numberAll;
                    }
                }

                bestNumber = number;
                finalObject = object;

                // is it source ?
                if (firstFound) {
                    isResultSource = false;
                } else {
                    isResultSource = true;
                }

            }
        }

        // nothing found
        if (bestNumber == -1) {
            return null;
        }

        // if found both of adress in results throw this exception
        // return found object
        if (isResultSource) {
            // switch the addresses
            String helpAddr = finalObject.getSourceAddress();
            finalObject.setSourceAddress(finalObject.getDestAddress());
            finalObject.setDestAddress(helpAddr);
        }

        return finalObject;
    }

    /**
     * Evaluates the value of distance by given strength. Higher strength -> lower distance.
     * If strength is 0 returns the max distance.
     * @param strength Strength between two motes.
     * @return Returns the distance.
     */
    public int evaluateDistanceFromStrength(int strength) {

        if (strength == 0) {
            return MAXDISTANCE;
        }

        //double distance = (1 - ((double) strength / 100)) * MAXDISTANCE;
        int distance = MAXDISTANCE - ((MAXDISTANCE * strength) / 100);

        return distance;
    }

    /**
     * It is used to get the borders of screen. It returns the positions of borders objects. The returned
     * value is extended by <code>PADDING</code> value.
     * @param objectList The List of all final determined objects.
     * @return Returns an instance of RectangleObject that presents borders of window.
     */
    public RectangleObject getWindowBorder(List<GeometryObject> objectList) {

        if (objectList == null) {
            return new RectangleObject(0, 0, 10, 10);
        }

        // size of the list
        int size = objectList.size();

        // init borders
        // if we want to get the real min/max of window by replacing the lower/higher value
        // we need to switch the initial max numbers.
        RectangleObject border = new RectangleObject(MAXXBORDER, MAXYBORDER, MINXBORDER, MINYBORDER);



        // go through all objects
        for (int i = 0; i < size; i++) {

            GeometryObject object = objectList.get(i);

            // if the object is not a point skip it
            if (object.getName().compareTo(GeometryObject.POINTNAME) != 0) {
                continue;
            }

            // get concrete object
            PointObject point = (PointObject) object;

            // if less replace new min value
            // -------------------------
            if (point.getX() < border.getX1()) {
                border.setX1(point.getX());
            }

            if (point.getY() < border.getY1()) {
                border.setY1(point.getY());
            }

            // if more replace new max value
            // -------------------------
            if (point.getX() > border.getX2()) {
                border.setX2(point.getX());
            }

            if (point.getY() > border.getY2()) {
                border.setY2(point.getY());
            }
        }


        // set padding
        border.setX1(border.getX1() - PADDING);
        border.setY1(border.getY1() - PADDING);
        border.setX2(border.getX2() + PADDING);
        border.setY2(border.getY2() + PADDING);

        // return border
        return border;
    }

    /**
     * This method gets all lines between motes which see each other.
     * @param resultPoints The List of final points of topology.
     * @param rssiObjectList The List of <sourceAddr, destAdr, strength> records.
     * @return Returns a list of lines. Each line connects two motes.
     */
    private List<GeometryObject> getLines(List<RSSIObject> rssiObjectList,
            List<GeometryObject> resultPoints) {

        // result
        List lines = new ArrayList<GeometryObject>();

        PointObject source = null;
        PointObject dest = null;


        int resultSize = resultPoints.size();
        int rssiObjectSize = rssiObjectList.size();


        // check if empty
        if (rssiObjectSize == 0) {
            return null;
        }


        // iterate the list of motes with determined position
        // -------------------------------------------
        for (int i = 0; i < rssiObjectSize; i++) {
            RSSIObject object = rssiObjectList.get(i);

            // iterate the results = determined points
            for (int j = 0; j < resultSize; j++) {
                PointObject pointObject = (PointObject) resultPoints.get(j);
                // if it is not Point continue
                if (pointObject.getName().compareTo(GeometryObject.POINTNAME) != 0) {
                    continue;
                }

                // found source point
                if (pointObject.getId().compareTo(object.getSourceAddress()) == 0) {
                    source = pointObject;
                    continue;
                }

                // found dest point
                if (pointObject.getId().compareTo(object.getDestAddress()) == 0) {
                    dest = pointObject;
                    continue;
                }
            }


            // if error
            if (source == null || dest == null) {
                break;
            }

            // found both
            // add line
            LineObject line = new LineObject(source.getX(), source.getY(),
                    dest.getX(), dest.getY(), object.getStrength());
            lines.add(line);
        }
        // -----------------------------------------


        // return the list of lines
        return lines;
    }
}
