// -----------------------------------------------------------
// TraceDrawer.java
// -----------------------------------------------------------

package cz.vutbr.wsagent.util;
import java.awt.Color;
import java.awt.image.BufferedImage;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.util.List;
import java.io.IOException;
import javax.imageio.ImageIO;
import java.io.File;
import java.awt.Font;
import java.awt.Image;
import java.awt.BasicStroke;

/**
 * Class for display agent movement.
 * @author Marek Houšť
 */
public class TraceDrawer {

    /* COLOR PROPERTIES */
    // backgroundColor
    private static final Color bgColor = new Color(0xFB,0xF7,0xF7);
    // color of text beside nodes 6c8395
    private static final Color nodeTextColor = new Color(0x77,0x00,0x00);
    // color of text move agent
    private static final Color agentMoveColor = new Color(0x77,0x00,0x00);
    // color of trace line between larger and smaller mote adress
    private static final Color traceLineColor = new Color(0x00,0x00,0x00);
    // thick of trace line
    private static final float traceLineThick = 3.0f;

    private BufferedImage bufferedImage;
    private String webPath;


    /**
     * Constructor for creating new instance. It prepares image to draw in it.
     * @param width The width of scene.
     * @param height The heightof scene.
     * @param inputPath The path where should be all images saved.
     */
    public TraceDrawer(int width, int height, String inputPath){

        // set the path of images
        webPath = inputPath;

        // create new buffered image
        bufferedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);

        // creating graphics
        Graphics g = bufferedImage.getGraphics();
        Graphics2D g2d = (Graphics2D) g;

        // formatting graphic properties
        g2d.setBackground(bgColor);
        g2d.clearRect(0, 0, width, height);

        g2d.setColor(bgColor);
        g2d.fillRect(0, 0, width, height);

    }

    /**
     * Calculated objects are simply drawn.
     * @param trackingAgentObjects
     * @param border
     * @throws IOException
     */
    public void drawTraces(List<GeometryObject> trackingAgentObjects, RectangleObject border)  throws IOException {


        for(int i = 0; i < trackingAgentObjects.size(); i++) {
            GeometryObject geoObject = (GeometryObject) trackingAgentObjects.get(i);

            // if it's point object - it is mote without agent
            if(geoObject.getName().compareTo(GeometryObject.POINTNAME) == 0) {
                try {
                    PointObject point = (PointObject) geoObject;
                    // draw mote without agent
                    point.setX(point.getX()-border.getX1());
                    point.setY(point.getY()-border.getY1());
                    drawMoteWithoutAgent(point);
                } catch (IOException ex) {
                }
            }

            // if it's trace point object - it is mote with agent
            if(geoObject.getName().compareTo(GeometryObject.TRACEPOINT) == 0) {
                try {
                    TracePoint point = (TracePoint) geoObject;
                    // draw mote with agent
                    point.setX(point.getX()-border.getX1());
                    point.setY(point.getY()-border.getY1());
                    drawMoteWithAgent(point);
                } catch (IOException ex) {
                }
            }

            // if it is trace line
            if(geoObject.getName().compareTo(GeometryObject.TRACELINE) == 0) {
                try {
                    TraceLine line = (TraceLine) geoObject;
                    // draw trace line
                    line.setX1(line.getX1()-border.getX1());
                    line.setY1(line.getY1()-border.getY1());
                    line.setX2(line.getX2()-border.getX1());
                    line.setY2(line.getY2()-border.getY1());
                    drawTraceLine(line);
                } catch (IOException ex) {
                }
            }

            // if it is trace line
            if(geoObject.getName().compareTo(GeometryObject.AGENTPOINT) == 0) {
                try {
                    AgentPoint point = (AgentPoint) geoObject;
                    // draw agent point (depends on the coordinates of the trace line)
                    point.setX(point.getX()-border.getX1());
                    point.setY(point.getY()-border.getY1());
                    drawAgentPoint(point);
                } catch (IOException ex) {
                }
            }

        }

    }

    /**
     * This method is for drawing topology image.
     * @param pathFile the file.
     * @exception IOException Thrown when the given path is incorrect.
     */
    public void generateTopologyImage(String pathFile) throws IOException {

        // creating graphics
        Graphics g = bufferedImage.getGraphics();
        Graphics2D g2d = (Graphics2D) g;

        g2d.dispose();

        // writes the image to the disk of given path
        ImageIO.write(bufferedImage, "jpg", new File(webPath+pathFile));
    }

    /**
     * Get the value of webPath
     *
     * @return the value of webPath
     */
    public String getWebPath() {
        return webPath;
    }

    /**
     * Set the value of webPath
     *
     * @param webPath new value of webPath
     */
    public void setWebPath(String webPath) {
        this.webPath = webPath;
    }


    /**
     * Get the value of bufferedImage
     *
     * @return the value of bufferedImage
     */
    public BufferedImage getBufferedImage() {
        return bufferedImage;
    }

    /**
     * Set the value of bufferedImage
     *
     * @param bufferedImage new value of bufferedImage
     */
    public void setBufferedImage(BufferedImage bufferedImage) {
        this.bufferedImage = bufferedImage;
    }

    /**
     * Method draws mote with agent. Coordinates are the centre of object.
     * @param point The trace point object.
     */
    private void drawMoteWithAgent(TracePoint point) throws IOException{

        Graphics g = bufferedImage.getGraphics();
        Graphics2D g2d = (Graphics2D) g;

        // mote drawing
        String moteSourceFile = webPath+"/ground/images/topology/mote.jpg";
        File moteFile = new File(moteSourceFile);
        Image buffImage = ImageIO.read(moteFile);

        g2d.drawImage(buffImage, point.getX()-30, point.getY()-16, Color.red, null);

        int length;
        // text mote
        g2d.setFont(new Font("Goudy Handtooled BT", Font.BOLD, 12));
        g2d.setPaint(nodeTextColor);
        length = (point.getId()).length();
        g2d.drawString(point.getId(), point.getX()-(5*length), point.getY()+20);
        
    }

    /**
     * Method draws mote without agent. Coordinates are the centre of object.
     * @param point The point object.
     */
    private void drawMoteWithoutAgent(PointObject point) throws IOException{

        Graphics g = bufferedImage.getGraphics();
        Graphics2D g2d = (Graphics2D) g;

        // mote drawing
        String moteSourceFile = webPath+"/ground/images/topology/motewithoutagent.jpg";
        File moteFile = new File(moteSourceFile);
        Image buffImage = ImageIO.read(moteFile);

        g2d.drawImage(buffImage, point.getX()-30, point.getY()-16, Color.red, null);

        int length;
        // text mote
        g2d.setFont(new Font("Goudy Handtooled BT", Font.BOLD, 12));
        g2d.setPaint(nodeTextColor);
        length = (point.getId()).length();
        g2d.drawString(point.getId(), point.getX()-(5*length), point.getY()+20);
        
    }

    /**
     * Method draws line with given points.
     * @param line The line object.
     */
    private void drawTraceLine(TraceLine traceLine) throws IOException {

        Graphics g = bufferedImage.getGraphics();
        Graphics2D g2d = (Graphics2D) g;

        g2d.setColor(traceLineColor);
        g2d.setStroke(new BasicStroke(traceLineThick));
        
        drawArrow(g2d, traceLine.getX1(), traceLine.getY1(), traceLine.getX2(), traceLine.getY2());

    }

    /**
     * Method for drawing the counter movement.
     * @param agentPoint
     * @throws IOException
     */
    private void drawAgentPoint(AgentPoint agentPoint) throws IOException {
        
        List<Long> listCounters = agentPoint.getListCounter();
        int x = agentPoint.getX();
        int y = agentPoint.getY();

        Graphics g = bufferedImage.getGraphics();
        Graphics2D g2d = (Graphics2D) g;
        g2d.setFont(new Font("Goudy Handtooled BT", Font.BOLD, 12));
        g2d.setPaint(agentMoveColor);
        
        for (int i = 0; i < listCounters.size(); i++) {

            g2d.drawString("c=" + listCounters.get(i), x, y - i*13);
            
        }
    }

   /**
    * Draw arrow
    * @param g2d Graphics2D
    * @param x1 The x location of the "tail" of the arrow
    * @param y1 The y location of the "tail" of the arrow
    * @param x2 The x location of the "head" of the arrow
    * @param y2 The y location of the "head" of the arrow
    */
    private void drawArrow(Graphics2D g2d, int x1, int y1, int x2, int y2) {
        
        float arrowWidth = 10.0f ;
        float theta = 0.423f ;
        int[] xPoints = new int[3] ;
        int[] yPoints = new int[3] ;
        float[] vecLine = new float[2] ;
        float[] vecLeft = new float[2] ;
        float fLength;
        float th;
        float ta;
        float baseX, baseY ;
        
        xPoints[0] = x2 ;
        yPoints[0] = y2 ;

        // build the line vector
        vecLine[0] = (float)xPoints[0] - x1 ;
        vecLine[1] = (float)yPoints[0] - y1 ;
        
        // build the arrow base vector - normal to the line
        vecLeft[0] = -vecLine[1] ;
        vecLeft[1] = vecLine[0] ;

        // setup length parameters
        fLength = (float)Math.sqrt( vecLine[0] * vecLine[0] + vecLine[1] * vecLine[1] ) ;
        th = arrowWidth / ( 2.0f * fLength ) ;
        ta = arrowWidth / ( 2.0f * ( (float)Math.tan( theta ) / 2.0f ) * fLength ) ;

        // find the base of the arrow
        baseX = ( (float)xPoints[0] - ta * vecLine[0]);
        baseY = ( (float)yPoints[0] - ta * vecLine[1]);

        // build the points on the sides of the arrow
        xPoints[1] = (int)( baseX + th * vecLeft[0] );
        yPoints[1] = (int)( baseY + th * vecLeft[1] );
        xPoints[2] = (int)( baseX - th * vecLeft[0] );
        yPoints[2] = (int)( baseY - th * vecLeft[1] );
        
        g2d.drawLine( x1, y1, (int)baseX, (int)baseY ) ;
        g2d.fillPolygon( xPoints, yPoints, 3 ) ;
    }

}
